#include "cmdline.h"
#include "dirlisting.h"
#include "i4httools.h"
#include "request.h"
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

static void print_file (FILE *tx, FILE *file);
static void request (FILE *tx, char *path, char *addr);
static int filter_directory (const struct dirent *dirent);
static void print_directory (FILE *tx, char *full_path, char *addr);
static void exec_perl_script (FILE *tx, char *full_path, char *addr);
static void error_internally (FILE *tx, char *address, char *msg);

int
initRequestHandler (void)
{
  const char *path = cmdlineGetValueForKey ("wwwpath");
  if (path == NULL)
    {
      return -1;
    }

  DIR *dir = opendir (path);
  if (dir == NULL)
    {
      perror ("opendir");
      return -1;
    }
  if (closedir (dir) == -1)
    {
      perror ("closedir");
      return -1;
    }

  return 0;
}

void
handleRequest (FILE *rx, FILE *tx)
{
  int maxLength = 8193;
  char line[maxLength];
  if (fgets (line, maxLength, rx) == NULL)
    {
      if (ferror (rx))
        {
          error_internally (tx, NULL, "fgets");
          return;
        }
      else
        {
          fprintf (stderr, "Client didnt responded");
          httpBadRequest (tx);
          return;
        }
    }

  char *saveptr;
  char *get = strtok_r (line, " ", &saveptr);
  if (get == NULL)
    {
      httpBadRequest (tx);
      return;
    }
  if (strcmp (get, "GET") != 0)
    {
      httpBadRequest (tx);
      return;
    }

  char *address = strtok_r (NULL, " ", &saveptr);
  if (address == NULL)
    {
      httpBadRequest (tx);
      return;
    }

  char *end = strtok_r (NULL, " ", &saveptr);
  if (end == NULL)
    {
      httpBadRequest (tx);
      return;
    }

  if (strcmp (end, "HTTP/1.0\r\n") != 0 && strcmp (end, "HTTP/1.0\n") != 0
      && strcmp (end, "HTTP/1.1\r\n") != 0 && strcmp (end, "HTTP/1.1\n") != 0)
    {
      httpBadRequest (tx);
      return;
    }

  const char *path = cmdlineGetValueForKey ("wwwpath");
  if (path == NULL)
    {
      httpInternalServerError (tx, address);
      return;
    }

  char full_path[strlen (path) + strlen (address) + 1];

  strcpy (full_path, path);
  strcat (full_path, address);

  request (tx, full_path, address);
}

static void
request (FILE *tx, char *full_path, char *addr)
{
  if (checkPath (addr) == -1)
    {
      fprintf (stderr, "Illegal access detected %s\n", full_path);
      httpForbidden (tx, addr);
      return;
    }

  errno = 0;
  int file_fd = open (full_path, O_RDONLY | O_CLOEXEC);
  if (file_fd == -1)
    {
      perror ("open");
      if (errno == ENOENT)
        {
          httpNotFound (tx, addr);
        }
      else if (errno == EACCES)
        {
          httpForbidden (tx, addr);
        }
      else
        {
          httpInternalServerError (tx, addr);
        }
      return;
    }

  FILE *file = fdopen (file_fd, "r");
  if (file == NULL)
    {
      error_internally (tx, addr, "fdopen");
      close (file_fd);
      return;
    }

  struct stat statbuf;
  if (fstat (file_fd, &statbuf) == -1)
    {
      error_internally (tx, addr, "lstat");
      fclose (file);
      return;
    }

  if (S_ISDIR (statbuf.st_mode))
    {
      if (addr[strlen (addr) - 1] != '/')
        {
          char new_addr[strlen (addr) + 2];
          strcpy (new_addr, addr);
          new_addr[strlen (addr)] = '/';
          new_addr[strlen (addr) + 1] = '\0';
          httpMovedPermanently (tx, new_addr);
          return;
        }
      else
        {
          char index_path[strlen (full_path) + strlen ("index.html") + 1];
          strcpy (index_path, full_path);
          strcat (index_path, "index.html");

          char index_addr[strlen (addr) + strlen ("index.html") + 1];
          strcpy (index_addr, addr);
          strcat (index_addr, "index.html");

          if (access (index_path, F_OK) == 0)
            {
              request (tx, index_path, index_addr);
            }
          else
            {
              print_directory (tx, full_path, addr);
            }
        }
    }
  else if (S_ISREG (statbuf.st_mode))
    {
      if (strlen (addr) > 3
          && strncmp (addr + strlen (addr) - 3, ".pl", 3) == 0
          && statbuf.st_mode & S_IXUSR)
        {
          exec_perl_script (tx, full_path, addr);
        }
      else
        {
          print_file (tx, file);
        }
    }
  else
    {
      httpBadRequest (tx);
    }

  if (fflush (tx) == EOF)
    {
      perror ("fflush");
      return;
    }

  if (fclose (file) == EOF)
    {
      perror ("fclose");
      return;
    }
}

static void
print_file (FILE *tx, FILE *file)
{
  httpOK (tx);

  int c = 0;
  while (!feof (file))
    {
      c = fgetc (file);
      if (c == EOF)
        {
          break;
        }

      if (fputc (c, tx) == EOF)
        {
          perror ("fputc");
          return;
        }
    }

  if (ferror (file))
    {
      perror ("fgetc");
      return;
    }
}

static void
print_directory (FILE *tx, char *full_path, char *address)
{
  struct dirent **names;
  int size = scandir (full_path, &names, &filter_directory, &alphasort);
  if (size == -1)
    {
      error_internally (tx, address, "scandir");
      return;
    }

  httpOK (tx);

  dirlistingBegin (tx, address);
  for (int i = 0; i < size; i++)
    {
      dirlistingPrintEntry (tx, address, names[i]->d_name);
      free (names[i]);
    }
  dirlistingEnd (tx);
  free (names);
}

static int
filter_directory (const struct dirent *dirent)
{
  return dirent->d_name[0] != '.';
}

static void
exec_perl_script (FILE *tx, char *full_path, char *addr)
{
  pid_t pid = fork ();
  if (pid == -1)
    {
      error_internally (tx, addr, "fork");
      return;
    }
  else if (pid == 0)
    {
      if (fflush (stdout) == EOF)
        {
          error_internally (tx, addr, "fflush");
          return;
        }

      if (dup2 (fileno (tx), STDOUT_FILENO) == -1)
        {
          error_internally (tx, addr, "dup2");
          return;
        }

      execl (full_path, addr);
      error_internally (tx, addr, "execl");
      exit (EXIT_FAILURE);
    }
  else
    {
      int wstatus;
      while (waitpid (pid, &wstatus, 0) != -1)
        {
          if (WIFEXITED (wstatus) || WIFSIGNALED (wstatus))
            {
              return;
            }
        }
      perror ("waitpid");
      return;
    }
}

static void
error_internally (FILE *tx, char *address, char *msg)
{
  perror (msg);
  httpInternalServerError (tx, address);
}
