#include "queue.h"
#include "run.h"
#include "sem.h"
#include <errno.h>
#include <limits.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static void *execCallback (void *argv);
static void *printCallback (void *argv);

struct param
{
  SEM *sem;
  char *cmd;
};

static void
die (const char *s)
{
  perror (s);
  exit (EXIT_FAILURE);
}

static int
parse_positive_int_or_die (char *str)
{
  errno = 0;
  char *endptr;
  long x = strtol (str, &endptr, 10);
  if (errno != 0)
    {
      die ("invalid number");
    }
  // Non empty string was fully parsed
  if (str == endptr || *endptr != '\0')
    {
      fprintf (stderr, "invalid number\n");
      exit (EXIT_FAILURE);
    }
  if (x <= 0)
    {
      fprintf (stderr, "number not positive\n");
      exit (EXIT_FAILURE);
    }
  if (x > INT_MAX)
    {
      fprintf (stderr, "number too large\n");
      exit (EXIT_FAILURE);
    }
  return (int)x;
}

int
main (int argc, char **argv)
{
  if (argc != 3)
    {
      fprintf (stderr, "usage: ./mach <threads> <mach file>\n");
      exit (EXIT_FAILURE);
    }

  int threads = parse_positive_int_or_die (argv[1]);

  SEM *sem = semCreate (threads);
  if (sem == NULL)
    {
      die ("semCreate");
    }

  if (queue_init ())
    {
      die ("queue_init");
    }

  FILE *machfile = fopen (argv[2], "r");
  if (machfile == NULL)
    {
      die ("fopen");
    }

  pthread_t printThread;
  errno = pthread_create (&printThread, NULL, &printCallback, NULL);
  if (errno)
    {
      die ("pthread_create");
    }

  while (1)
    {
      char buffer[4097];
      if (fgets (buffer, 4097, machfile) == NULL)
        {
          if (ferror (machfile))
            {
              die ("fgets");
            }
          else
            {
              break;
            }
        }
      if (buffer[strlen (buffer) - 1] == '\n')
        {
          buffer[strlen (buffer) - 1] = '\0';
        }

      if (buffer[0] == '\0')
        {
          for (int j = 0; j < threads; j++)
            {
              P (sem);
            }
          for (int j = 0; j < threads; j++)
            {
              V (sem);
            }
        }
      else
        {
          char *cmd = strdup (buffer);
          if (cmd == NULL)
            {
              die ("strdup");
            }

          P (sem);

          struct param *param = malloc (sizeof (struct param));
          if (param == NULL)
            {
              die ("malloc");
            }
          param->cmd = cmd;
          param->sem = sem;

          pthread_t thread;
          errno = pthread_create (&thread, NULL, &execCallback, param);
          if (errno)
            {
              die ("pthread_create");
            }
        }
    }

  if (fclose (machfile) == EOF)
    {
      die ("fclose");
    }

  for (int j = 0; j < threads; j++)
    {
      P (sem);
    }

  if (queue_put (NULL, NULL, 0))
    {
      die ("queue_put");
    }

  errno = pthread_join (printThread, NULL);
  if (errno)
    {
      die ("pthread_join");
    }

  semDestroy (sem);
  queue_deinit ();

  return EXIT_SUCCESS;
}

static void *
printCallback (void *argv)
{
  while (1)
    {
      char *cmd;
      char *out;
      int flags;
      if (queue_get (&cmd, &out, &flags))
        {
          die ("queue_get");
        }

      if (cmd == NULL && out == NULL)
        {
          return EXIT_SUCCESS;
        }

      if (cmd != NULL && out == NULL)
        {
          fprintf (stdout, "Running `%s`\n", cmd);
        }
      else if (cmd != NULL && out != NULL)
        {
          fprintf (stdout, "Completed `%s` : \"%s\"\n", cmd, out);
          free (cmd);
          free (out);
        }

      if (fflush (stdout) == EOF)
        {
          die ("fflush");
        }
    }
}

static void *
execCallback (void *argv)
{
  errno = pthread_detach (pthread_self ());
  if (errno)
    {
      die ("pthread_detach");
    }

  struct param *param = (struct param *)argv;
  SEM *sem = param->sem;
  char *cmd = param->cmd;
  free (param);

  if (queue_put (cmd, NULL, 0))
    {
      die ("queue_put");
    }

  char *out;
  run_cmd (cmd, &out);

  if (queue_put (cmd, out, 0))
    {
      die ("queue_put");
    }

  V (sem);
  return EXIT_SUCCESS;
}
