#include "halde.h"
#include <errno.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <unistd.h>

/// Magic value for occupied memory chunks.
#define MAGIC ((void *)0xbffdfa0d)

/// Size of the heap (in bytes).
#define SIZE (4096 * 4096 * 1)

/// Memory-chunk structure.
struct mblock
{
  struct mblock *next;
  size_t size;
  char memory[];
};

/// Heap-memory area.
static char *memory;

/// Pointer to the first element of the free-memory list.
static struct mblock *first;

/// Helper function to visualise the current state of the free-memory list.
void
printList (void)
{
  struct mblock *lauf = first;

  // Empty list
  if (first == NULL)
    {
      char empty[] = "(empty)\n";
      write (STDERR_FILENO, empty, sizeof (empty));
      return;
    }

  // Print each element in the list
  const char fmt_init[] = "(off: %7zu, size:: %7zu)";
  const char fmt_next[] = " --> (off: %7zu, size:: %7zu)";
  const char *fmt = fmt_init;
  char buffer[sizeof (fmt_next) + 2 * 7];

  while (lauf)
    {
      size_t n = snprintf (buffer, sizeof (buffer), fmt,
                           (uintptr_t)lauf - (uintptr_t)memory, lauf->size);
      if (n)
        {
          write (STDERR_FILENO, buffer, n);
        }

      lauf = lauf->next;
      fmt = fmt_next;
    }
  write (STDERR_FILENO, "\n", 1);
}

void *
malloc (size_t size)
{
  if (size > SIZE)
    {
      errno = ENOMEM;
      return NULL;
    }

  if (memory == NULL)
    {
      memory = mmap (NULL, SIZE, PROT_READ | PROT_WRITE,
                     MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
      if (memory == MAP_FAILED)
        {
          memory = NULL;
          return NULL;
        }

      first = (struct mblock *)memory;
      first->next = NULL;
      first->size = SIZE - sizeof (struct mblock);
    }

  if (first == NULL)
    {
      errno = ENOMEM;
      return NULL;
    }

  struct mblock *previous = NULL;
  struct mblock *current = first;

  while (current != NULL && (current->size < size))
    {
      previous = current;
      current = current->next;
    }

  if (current == NULL)
    {
      errno = ENOMEM;
      return NULL;
    }

  struct mblock *new;

  int rest = ((int)current->size) - size - sizeof (struct mblock);
  if (rest > 0)
    {
      char *mem = current->memory;
      mem += size;
      new = (struct mblock *)mem;
      new->size = (size_t)rest;
      new->next = current->next;

      current->size = size;
    }
  else
    {
      new = current->next;
    }

  if (current == first)
    {
      first = new;
    }
  else
    {
      previous->next = new;
    }

  current->next = MAGIC;

  return current->memory;
}

void
free (void *ptr)
{
  if (ptr == NULL)
    {
      return;
    }

  char *memory2 = ptr;
  memory2 -= sizeof (struct mblock);
  struct mblock *block = (struct mblock *)memory2;

  if (block->next != MAGIC)
    {
      abort ();
    }

  block->next = first;
  first = block;
}

void *
realloc (void *ptr, size_t size)
{
  if (ptr == NULL)
    {
      return malloc (size);
    }

  if (size == 0)
    {
      free (ptr);
      return NULL;
    }

  char *newPtr = malloc (size);
  if (newPtr == NULL)
    {
      return NULL;
    }

  struct mblock *old
      = (struct mblock *)(((char *)ptr) - sizeof (struct mblock));

  size_t smaller;
  if (size < old->size)
    {
      smaller = size;
    }
  else
    {
      smaller = old->size;
    }

  memcpy (newPtr, ptr, smaller);

  free (ptr);

  return newPtr;
}

void *
calloc (size_t nmemb, size_t size)
{
  if (nmemb == 0 || size == 0)
    {
      return NULL;
    }

  size_t sumSize = nmemb * size;
  if (sumSize / nmemb != size)
    {
      errno = ENOMEM;
      return NULL;
    }

  void *ptr = malloc (sumSize);
  if (ptr == NULL)
    {
      return NULL;
    }

  return memset (ptr, 0, sumSize);
}
